# -*- coding: utf-8 -*-
"""
Minimal qualitative test functionality. Compares outputs of the code to default outputs,
for the 3 reference input files. The judgement about whether the resulting changes are
as expected or not is left to the developers
"""

#%% Import required modules
import pandas as pd
import matplotlib.pyplot as plt
import os
#%% Function to test the output against reference results
'''
By default, reference results are provided for a series of 30 days and for
all th 3 reference input files. If this is changed, the function below must
be called for the correct number of input files and days through the related
parameters
'''

def test_output(results_folder,test_folder, num_input_files=3, num_days=30):
    
    def series_to_average(profile_series, num_days):
        
        average_series = profile_series[0:1440]
        for d in range(1,num_days):
            average_series = average_series + profile_series[0+1440*d:1440+1440*d].set_index(average_series.index)
        average_series = average_series/ (d+1)
        
        return average_series
    
    default_out = {}
    current_out = {}
    axes = {}
    
    fig = plt.figure(figsize=(14,14))
    
    for n in range(1,4):
        
        if n < num_input_files+1:
            default_out[n] = series_to_average(pd.read_csv('{}/output_file_{}.csv'.format(test_folder,n), index_col=0), num_days)
            current_out[n] = series_to_average(pd.read_csv('{}/output_file_{}.csv'.format(results_folder,n), index_col=0), num_days)
            
            axes[n] = fig.add_subplot((310+n))
       
            axes[n].plot(default_out[n], color='red', linewidth=2, label='default')
            axes[n].plot(current_out[n], color='blue',linewidth=1, alpha=0.8, linestyle='--', label='new results')
            axes[n].set_xmargin(0)
            axes[n].set_ymargin(0)
            axes[n].set_ylabel('Power demand (kW)')
        else:
            axes[n] = fig.add_subplot((310+n))
            axes[n].set_xmargin(0)
            axes[n].set_ymargin(0)

    axes[n].get_shared_x_axes().join(axes[n], axes[n-1], axes[n-2])
    axes[n-1].legend()
    axes[n-1].set_xticklabels([])
    axes[n-2].set_xticklabels([])
    
#%% Testing the output and providing visual result
'''
Here the visual comparison between default and new/current results occurs.
Besides the difference naturally occurring due to different realisations of stochastic
parameters, the developers should check whether any other differences are brought by
by the tested code changes. If any differences are there, the developers should 
evaluate whether these are as expected/designed or not
'''

test_output('../results','../test', num_input_files=3)


# -*- coding: utf-8 -*-
"""
Created on Fri Apr 19 14:35:00 2019
This is the code for the open-source stochastic model for the generation of 
multi-energy load profiles in off-grid areas, called RAMP, v0.3.0.

@authors:
- Francesco Lombardi, Politecnico di Milano
- Sergio Balderrama, Université de Liège
- Sylvain Quoilin, KU Leuven
- Emanuela Colombo, Politecnico di Milano

Copyright 2019 RAMP, contributors listed above.
Licensed under the European Union Public Licence (EUPL), Version 1.2;
you may not use this file except in compliance with the License.

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and limitations
under the License.
"""

#%% Import required modules

import sys,os

import numpy as np

sys.path.append('../')

try:
    from .core.utils import get_day_type, yearly_pattern
    from .core.stochastic_process import stochastic_process
    from .post_process import post_process as pp
except ImportError:
    from core.utils import get_day_type, yearly_pattern
    from core.stochastic_process import stochastic_process
    from post_process import post_process as pp


def run_usecase(j=None, fname=None, num_profiles=None, days=None, plot=True, parallel=False):
    # Calls the stochastic process and saves the result in a list of stochastic profiles
    if days is None:
        Profiles_list = stochastic_process(j=j, fname=fname, num_profiles=num_profiles, day_type=yearly_pattern(), parallel=parallel)

        # Post-processes the results and generates plots
        Profiles_avg, Profiles_list_kW, Profiles_series = pp.Profile_formatting(Profiles_list)
        pp.Profile_series_plot(Profiles_series)  # by default, profiles are plotted as a series

        pp.export_series(Profiles_series, j, fname)

        if len(Profiles_list) > 1:  # if more than one daily profile is generated, also cloud plots are shown
            pp.Profile_cloud_plot(Profiles_list, Profiles_avg)
    else:
        Profiles_list = []
        if parallel is True:
            Profiles_list = stochastic_process(j=j, fname=fname, num_profiles=len(days), day_type=[get_day_type(day)for day in days], parallel=parallel)
        else:
            for day in days:
                print("Day", day)
                daily_profiles = stochastic_process(j=j, fname=fname, num_profiles=num_profiles, day_type=get_day_type(day), parallel=parallel)

                Profiles_list.append(np.mean(daily_profiles, axis=0))
        if plot is True:
            # Post-processes the results and generates plots
            Profiles_avg, Profiles_list_kW, Profiles_series = pp.Profile_formatting(Profiles_list)
            pp.Profile_series_plot(Profiles_series)  # by default, profiles are plotted as a series

            pp.export_series(Profiles_series, j, fname)

            if len(Profiles_list) > 1:  # if more than one daily profile is generated, also cloud plots are shown
                pp.Profile_cloud_plot(Profiles_list, Profiles_avg)
        else:
            return Profiles_list


input_files_to_run = [1, 2, 3]


if __name__ == "__main__":

    for j in input_files_to_run:
        try:
            run_usecase(j=j, fname='../example/input_file_{}.xlsx'.format(j))
        except:
            print('Input files in .xlsx format not found. Running the default files in .py format.')
            run_usecase(j=j)
            


# -*- coding: utf-8 -*-
"""
The code is based on UseCase, User and Appliance classes.
A UseCase instance consists of a list of User instances which own Appliance instances
Within the Appliance class, some other functions are created to define windows of use and,
if needed, specific duty cycles
"""
import numpy as np
import pandas as pd
from tqdm import tqdm
import multiprocessing
import warnings
import random
import math
from ramp.core.constants import NEW_TO_OLD_MAPPING, APPLIANCE_ATTRIBUTES, APPLIANCE_ARGS, WINDOWS_PARAMETERS, DUTY_CYCLE_PARAMETERS, switch_on_parameters
from ramp.core.utils import random_variation, duty_cycle, random_choice, read_input_file, within_peak_time_window


from typing import List, Union,Iterable
from ramp.errors_logs.errors import InvalidType,InvalidWindow

def single_appliance_daily_load_profile(args):
    app, args = args
    app.generate_load_profile(*args, power=app.power[args[0]])

    return args[0], app.daily_use


class UseCase:
    def __init__(self, name:str="", users:Union[List,None]=None):
        """ Creates a UseCase instance for gathering a list of User instances which own Appliance instances

        Parameters
        ----------
        name : str, optional
            name of the usecase instance, by default ""
        users : Union[Iterable,None], optional
            a list of users to be added to the usecase instance, by default None
        """
        self.name = name
        self.appliances = []
        if users is None:
            users = []
        self.users = users
        self.collect_appliances_from_users()

    def add_user(self, user) -> None:
        """adds new user to the user property list

        Parameters
        ----------
        user : User
            an user instance

        Raises
        ------
        InvalidType
            any type rather than User will raise the error.
        """
        if isinstance(user, User):
            self.users.append(user)
        else:
            raise InvalidType(f"{type(user)} is not valid. Only 'User' type is acceptable.")

    def collect_appliances_from_users(self):
        appliances = []
        for user in self.users:
            appliances = appliances + user.App_list
        self.appliances = appliances


    def generate_daily_load_profiles(self, num_profiles, peak_time_range, day_types):
        profiles = []
        for prof_i in range(num_profiles):
            # initialise an empty daily profile (or profile load)
            # that will be filled with the sum of the daily profiles of each User instance
            usecase_load = np.zeros(1440)
            # for each User instance generate a load profile, iterating through all user of this instance and
            # all appliances they own, corresponds to step 2. of [1], p.7
            for user in self.users:
                user.generate_aggregated_load_profile(prof_i, peak_time_range, day_types)
                # aggregate the user load to the usecase load
                usecase_load = usecase_load + user.load
            profiles.append(usecase_load)
            # screen update about progress of computation
            #print('Profile', prof_i+1, '/', num_profiles, 'completed')
        return profiles

    def generate_daily_load_profiles_parallel(self, num_profiles, peak_time_range, day_types):
        max_parallel_processes = multiprocessing.cpu_count()
        tasks = []
        t = 0
        for day_id in range(num_profiles):
            day_type = day_types[day_id]
            for user in self.users:
                for app in user.App_list:
                    for _ in range(user.num_users):
                        t = t + 1
                        tasks.append((app, (day_id, peak_time_range, day_type)))

        daily_profiles_dict = {}
        timeout = 1

        with multiprocessing.Pool(max_parallel_processes) as pool:
            with tqdm(
                    total=len(tasks),
                    desc=f"Computing appliances profiles",
                    unit="unit",
            ) as pbar:

                imap_unordered_it = pool.imap_unordered(single_appliance_daily_load_profile, tasks, chunksize=4)
                for prof_i, daily_load in imap_unordered_it:
                    if prof_i in daily_profiles_dict:
                        daily_profiles_dict[prof_i].append(daily_load)
                    else:
                        daily_profiles_dict[prof_i] = [daily_load]
                    pbar.update()

            daily_profiles = np.zeros((num_profiles, 1440))

            for day_id in range(num_profiles):
                daily_profiles[day_id, :] = np.vstack(daily_profiles_dict[day_id]).sum(axis=0)

        return daily_profiles

    def save(self, filename:str=None) -> Union[pd.DataFrame,None]:
        """Saves/returns the model databas including all the users and their appliances as a single pd.DataFrame or excel file.

        Parameters
        ----------
        filename : str, optional
            The path where the data will be stored. if None, function will return a pd.DataFrame, by default None.

        Returns
        -------
        Union[pd.DataFrame,None]
            if filename is passed, returnes None, otherwise will return a pd.DataFrame

        Notes
        -------
        The 'filename' parameter should consist the path and the name of the file without the file extension.
        For example, if the user wants to save database into "new_folder" directory and name the file as "ramp_database.xlsx" should use:

        .. code-block:: python

            usecase.save("new_folder/ramp_database")
        """
        answer = pd.concat([user.save() for user in self.users], ignore_index=True)
        if filename is not None:
            answer.to_excel(f"{filename}.xlsx", index=False, engine="openpyxl")
        else:
            return answer

    def export_to_dataframe(self) -> pd.DataFrame:
        """exports the model database to a pd.DataFrame containing all the data related to users and their appliances

        Returns
        -------
        pd.DataFrame
            the model database
        """
        return self.save()

    def load(self, filename:str) -> None:
        """Open an .xlsx file which was produced via the save method and create instances of Users and Appliances

        Parameters
        ----------
        filename : str
            The path where the excel database is.

        Raises
        ---------
        ValueError
            #. if the 'num_users' is not the same for a given user profile
            #. if the 'user_preference' is not the same for a given user profile

        Note
        ------
        The 'filename' parameter should consist the path and the name of the file with the file extension.
        For example, if the user wants to save database into "new_folder" directory and name the file as "ramp_database.xlsx" should use:

        .. code-block:: python

            usecase.save("new_folder/ramp_database.xlsx")
        """

        df = read_input_file(filename=filename)
        for user_name in df.user_name.unique():
            user_df = df.loc[df.user_name == user_name]
            num_users = user_df.num_users.unique()
            if len(num_users) == 1:
                num_users = num_users[0]
            else:
                raise ValueError(
                    "'num_users' should be the same for a given user profile"
                )
            user_preference = user_df.user_preference.unique()
            if len(user_preference) == 1:
                user_preference = user_preference[0]
            else:
                raise ValueError(
                    "'user_preference' should be the same for a given user profile"
                )

            # create user and add it to usecase
            user = User(user_name, num_users, user_preference)
            self.add_user(user)
            # itereate through the lines of the DataFrame, each line representing one Appliance instance
            for row in user_df.loc[
                :, ~user_df.columns.isin(["user_name", "num_users", "user_preference"])
            ].to_dict(orient="records"):
                # assign Appliance arguments
                appliance_parameters = {k: row[k] for k in APPLIANCE_ARGS}

                # assign windows arguments
                for k in WINDOWS_PARAMETERS:
                    if "window" in k:
                        w_start = row.get(k + "_start", np.NaN)
                        w_end = row.get(k + "_end", np.NaN)
                        if not np.isnan(w_start) and not np.isnan(w_end):
                            appliance_parameters[k] = np.array(
                                [w_start, w_end], dtype=np.intc
                            )
                    else:
                        val = row.get(k, np.NaN)
                        if not np.isnan(val):
                            appliance_parameters[k] = val

                # assign duty cycles arguments
                for duty_cycle_params in DUTY_CYCLE_PARAMETERS:
                    for k in duty_cycle_params:
                        if "cw" in k:
                            cw_start = row.get(k + "_start", np.NaN)
                            cw_end = row.get(k + "_end", np.NaN)
                            if not np.isnan(cw_start) and not np.isnan(cw_end):
                                appliance_parameters[k] = np.array(
                                    [cw_start, cw_end], dtype=np.intc
                                )
                        else:
                            val = row.get(k, np.NaN)
                            if not np.isnan(val):
                                appliance_parameters[k] = val

                user.add_appliance(**appliance_parameters)

        self.collect_appliances_from_users()





class User:
    def __init__(self, user_name:str="", num_users:int=1, user_preference:int=0):
        """Creates a User instance (User Category)

        Parameters
        ----------
        user_name : str, optional
            name of the user type, by default ""
        num_users : int, optional
            number of users within the resprective user-type, by default 1
        user_preference : int {0,1,2,3}, optional
            Related to cooking behaviour, how many types of meal a user wants a day (number of user preferences has to be defined here and will be further specified with pref_index parameter), by default 0
        """
        self.user_name = user_name
        self.num_users = num_users
        self.user_preference = user_preference
        self.load = None
        self.App_list = []  # each instance of User (i.e. each user class) has its own list of Appliances

    def __repr__(self):
        return self.save()[["user_name","num_users","name","number","power"]].to_string()

    def add_appliance(self, *args, **kwargs):
        """adds an appliance to the user category with all the appliance characteristics in a single function


        Returns
        -------
        Appliance
            returns the appliance instance
        """

        # parse the args into the kwargs
        if len(args) > 0:
            for a_name, a_val in zip(APPLIANCE_ARGS, args):
                kwargs[a_name] = a_val

        # collects windows arguments
        windows_args = {}
        for k in WINDOWS_PARAMETERS:
            if k in kwargs:
                windows_args[k] = kwargs.pop(k)

        # collects duty cycles arguments
        duty_cycle_parameters = {}
        for i, duty_cycle_params in enumerate(DUTY_CYCLE_PARAMETERS):
            cycle_parameters = {}
            for k in duty_cycle_params:
                if k in kwargs:
                    cycle_parameters[k] = kwargs.pop(k)
            if cycle_parameters:
                duty_cycle_parameters[i+1] = cycle_parameters

        app = Appliance(self, **kwargs)

        if windows_args:
            app.windows(**windows_args)
        for i in duty_cycle_parameters:
            app.specific_cycle(i, **duty_cycle_parameters[i])

        return app

    @property
    def maximum_profile(self) -> np.array:
        """Aggregate the theoretical maximal profiles of each appliance of the user by switching the appliance always on

        Returns
        --------
        np.array
        """
        user_max_profile = np.zeros(1440)
        for appliance in self.App_list:
            # Calculate windows curve, i.e. the theoretical maximum curve that can be obtained, for each app, by switching-on always all the 'n' apps altogether in any time-step of the functioning windows
            app_max_profile = appliance.maximum_profile  # this computes the curve for the specific App
            user_max_profile = np.vstack([user_max_profile, app_max_profile])  # this stacks the specific App curve in an overall curve comprising all the Apps within a User class
        return np.transpose(np.sum(user_max_profile, axis=0)) * self.num_users

    def save(self, filename:str=None) -> Union[pd.DataFrame,None]:
        """Saves/returns the model databas including allappliances as a single pd.DataFrame or excel file.

        Parameters
        ----------
        filename : str, optional
            The path where the data will be stored. if None, function will return a pd.DataFrame, by default None.

        Returns
        -------
        Union[pd.DataFrame,None]
            if filename is passed, returnes None, otherwise will return a pd.DataFrame

        Raises
        ------
        Exception
            if now appliaces is assigned to the user and the function is called.

        Notes
        -------
        1. The 'filename' parameter should consist the path and the name of the file without the file extension.
        For example, if the user wants to save database into "new_folder" directory and name the file as "ramp_database.xlsx" should use:

        .. code-block:: python

            user.save("new_folder/ramp_database")

        2. Appliances are added to the user-type only if **'windows'** method of the Appliance is called.
        """

        try:
            answer = pd.concat([app.save() for app in self.App_list], ignore_index=True)
        except ValueError:
            raise Exception("No appliances is assigned to the user.")

        if filename is not None:
            answer.to_excel(f"{filename}.xlsx", engine="openpyxl")
        else:
            return answer

    def __eq__(self, other_user):
        """Compare two users

        ensure they have the same properties
        ensure they have the same appliances
        """
        answer = np.array([])
        for attribute in ("user_name", "num_users", "user_preference"):
            if hasattr(self, attribute) and hasattr(other_user, attribute):
                np.append(
                    answer, [getattr(self, attribute) == getattr(other_user, attribute)]
                )
            else:
                print(f"Problem with {attribute} of user")
                np.append(answer, False)
        answer = answer.all()

        if answer is True:
            # user attributes match, continue to compare each appliance
            if len(self.App_list) == len(other_user.App_list):
                answer = np.array([])
                for my_appliance, their_appliance in zip(
                    self.App_list, other_user.App_list
                ):
                    temp = my_appliance == their_appliance
                    answer = np.append(answer, temp)
                if len(answer) > 0:
                    answer = answer.all()
                else:

                    if len(self.App_list) > 0:
                        answer = False
                    else:
                        # both users have no appliances
                        answer = True
            else:
                print(
                    f"The user {self.user_name} and {other_user.user_name} do not have the same number of appliances"
                )
                answer = False
        return answer

    def export_to_dataframe(self) -> pd.DataFrame:
        """Saves/returns the model databas including allappliances as a single pd.DataFrame or excel file.

        Returns
        -------
        pd.DataFrame
            if filename is passed, returnes None, otherwise will return a pd.DataFrame

        Raises
        ------
        Exception
            if now appliaces is assigned to the user and the function is called.

        Notes
        -------
        Appliances are added to the user-type only if **'windows'** method of the Appliance is called.
        """
        return self.save()


    def Appliance(
        self,
        number=1,
        power=0,
        num_windows=1,
        func_time=0,
        time_fraction_random_variability=0,
        func_cycle=1,
        fixed="no",
        fixed_cycle=0,
        occasional_use=1,
        flat="no",
        thermal_P_var=0,
        pref_index=0,
        wd_we_type=2,
        name="",
    ):
        """Back-compatibility with legacy code

        Notes
        ------
        refer to Appliance class docs
        """
        return self.add_appliance(
            number=number,
            power=power,
            num_windows=num_windows,
            func_time=func_time,
            time_fraction_random_variability=time_fraction_random_variability,
            func_cycle=func_cycle,
            fixed=fixed,
            fixed_cycle=fixed_cycle,
            occasional_use=occasional_use,
            flat=flat,
            thermal_p_var=thermal_P_var,
            pref_index=pref_index,
            wd_we_type=wd_we_type,
            name=name,
        )

    def generate_single_load_profile(self, prof_i:int, peak_time_range:np.array, day_type:int):
        """Generates a load profile for a single user taking all its appliances into consideration

        Parameters
        ----------
        prof_i: int[0,365]
            ith profile (day) requested by the user. 0 is the first day of the year and 364 is the last day.

        peak_time_range: np.array
            randomised peak time range calculated using calc_peak_time_range function.

        day_type: int[0,1]
            type of the ith profile. 0 for a week day or 1 for a weekend day

        Returns
        --------
        np.array
            load profile for the requested day
        """

        if prof_i not in range(365):
            raise ValueError(f'prof_i should be an integer in range of 0 to 364')

        single_load = np.zeros(1440)


        for App in self.App_list:  # iterates for all the App types in the given User class

            App.generate_load_profile(prof_i, peak_time_range, day_type, power=App.power[prof_i])

            single_load = single_load + App.daily_use  # adds the Appliance load profile to the single User load profile
        return single_load

    def generate_aggregated_load_profile(self, prof_i, peak_time_range, day_type):
        """Generates an aggregated load profile from single load profile of each user


        Parameters
        ----------

        prof_i: int[0,365]
            ith profile (day) requested by the user. 0 is the first day of the year and 364 is the last day.
        peak_time_range: numpy array
            randomised peak time range calculated using calc_peak_time_range function
        day_type: int[0,1]
            type of the ith profile. 0 for a week day or 1 for a weekend day

        Returns
        --------
        np.array
            load profile for the requested day

        Notes
        ------
        Each single load profile has its own separate randomisation
        """

        if prof_i not in range(365):
            raise ValueError(f'prof_i should be an integer in range of 0 to 364')


        self.load = np.zeros(1440)  # initialise empty load for User instance
        for _ in range(self.num_users):
            # iterates for every single user within a User class.
            self.load = self.load + self.generate_single_load_profile(prof_i, peak_time_range, day_type)

        return self.load

class Appliance:
    def __init__(
        self,
        user,
        number:int=1,
        power:Union[float,pd.DataFrame]=0,
        num_windows:int=1,
        func_time:int=0,
        time_fraction_random_variability:float=0,
        func_cycle:int=1,
        fixed:str="no",
        fixed_cycle:int=0,
        occasional_use:float=1,
        flat:str="no",
        thermal_p_var:int=0,
        pref_index:int=0,
        wd_we_type:int=2,
        name:str="",
    ):
        """Creates an appliance for a given user

        Parameters
        ----------
        user : ramp.User
            user to which the appliance is bounded

        number : int, optional
            number of appliances of the specified kind, by default 1

        power : Union[float.pd.DataFrame], optional
            Power rating of appliance (average). If the appliance has variant daily power, a series (with the size of 365) can be passed., by default 0

        num_windows : int [1,2,3], optional
            Number of distinct time windows, by default 1

        func_time : int[0,1440], optional
            total time (minutes) the appliance is on during the day (not dependant on windows). Acceptable values are in range 0 to 1440, by default 0

        time_fraction_random_variability : Percentage, optional
            percentage of total time of use that is subject to random variability. For time (not for windows), randomizes the total time the appliance is on, by default 0

        func_cycle : int[0,1440], optional
            minimum time(minutes) the appliance is kept on after switch-on event, by default 1

        fixed : str, optional
            if 'yes', all the 'n' appliances of this kind are always switched-on together, by default "no"

        fixed_cycle : int{0,1,2,3,4}, optional
            Number of duty cycle, 0 means continuous power, if not 0 you have to fill the cw (cycle window) parameter (you may define up to 3 cws), by default 0

        occasional_use : Percentage, optional
            Defines how often the appliance is used, e.g. every second day will be 0.5, by default 1

        flat : str{'yes','no'}, optional
            allows to model appliances that are not subject to any kind of random variability, such as public lighting, by default "no"

        thermal_p_var : Percentage, optional
            Range of change of the power of the appliance (e.g. shower not taken at same temparature) or for the power of duty cycles (e.g. for a cooker, AC, heater if external temperature is different…), by default 0

        pref_index : int{0,1,2,3}, optional
            defines preference index for association with random User daily preference behaviour.This number must be smaller or equal to the value input in user_preference, by default 0

        wd_we_type : int{0,1,2}, optional
            Specify whether the appliance is used only on weekdays (0), weekend (1) or the whole week (2), by default 2

        name : str, optional
            the name of the appliance, by default ""

        Raises
        --------
        ValueError
            1. if power is not passed as a number of series.
            2. power array size is not (365,1)
        """

        self.user = user
        self.name = name
        self.number = number
        self.num_windows = num_windows
        self.func_time = func_time
        self.time_fraction_random_variability = time_fraction_random_variability
        self.func_cycle = (
            func_cycle
        )
        self.fixed = fixed
        self.fixed_cycle = fixed_cycle
        self.occasional_use = occasional_use
        self.flat = flat
        self.thermal_p_var = (
            thermal_p_var
        )
        self.pref_index = pref_index
        self.wd_we_type = wd_we_type

        if isinstance(power,pd.DataFrame):
            if power.shape == (365,1):
                power = power.values[:,0]
            else:
                raise ValueError("wrong size of array. array size should be (365,1).")

        elif isinstance(power,str):
            power = pd.read_json(power).values[:,0]

        elif isinstance(power,(float,int)):
            # TODO change this automatic value depending on the range of the usecase
            power = power * np.ones(366)

        else:
            raise ValueError("wrong data type for power.")

        self.power = power

        # attributes initialized by self.windows
        self.random_var_w = 0
        self.window_1 = np.array([0, 0])
        self.window_2 = np.array([0, 0])
        self.window_3 = np.array([0, 0])
        self.random_var_1 = 0
        self.random_var_2 = 0
        self.random_var_3 = 0
        self.daily_use = None
        self.free_spots = None

        # attributes used for specific fixed and random cycles
        self.p_11 = 0
        self.p_12 = 0
        self.t_11 = 0
        self.t_12 = 0
        self.r_c1 = 0
        self.p_21 = 0
        self.p_22 = 0
        self.t_21 = 0
        self.t_22 = 0
        self.r_c2 = 0
        self.p_31 = 0
        self.p_32 = 0
        self.t_31 = 0
        self.t_32 = 0
        self.r_c3 = 0

        # attribute used for cycle_behaviour
        self.cw11 = np.array([0, 0])
        self.cw12 = np.array([0, 0])
        self.cw21 = np.array([0, 0])
        self.cw22 = np.array([0, 0])
        self.cw31 = np.array([0, 0])
        self.cw32 = np.array([0, 0])

        self.random_cycle1 = np.array([])
        self.random_cycle2 = np.array([])
        self.random_cycle3 = np.array([])

    def save(self) -> pd.DataFrame:
        """returns a pd.DataFrame containing the appliance data

        Returns
        -------
        pd.DataFrame
            includes all the attributes and the user related information of an appliance.
        """
        dm = {}
        for user_attribute in ("user_name", "num_users", "user_preference"):
            dm[user_attribute] = getattr(self.user, user_attribute)
        for attribute in APPLIANCE_ATTRIBUTES:

            if hasattr(self, attribute):
                if "window_" in attribute or "cw" in attribute:
                    window_value = getattr(self, attribute)
                    dm[attribute + "_start"] = window_value[0]
                    dm[attribute + "_end"] = window_value[1]
                elif attribute == "power":
                    power_values = getattr(self, attribute)
                    if np.diff(power_values).sum() == 0:
                        power_values = power_values[0]
                    else:
                        power_values = power_values.tolist()
                    dm[attribute] = power_values
                else:
                    dm[attribute] = getattr(self, attribute)
            else:
                # this is for legacy purpose, so that people can export their old models to new format
                old_attribute = NEW_TO_OLD_MAPPING.get(attribute,attribute)
                if hasattr(self, old_attribute):
                    if "window_" in attribute or "cw" in attribute:
                        window_value = getattr(self, old_attribute)
                        dm[attribute + "_start"] = window_value[0]
                        dm[attribute + "_end"] = window_value[1]
                    elif old_attribute == "POWER":
                        power_values = getattr(self, old_attribute)
                        if np.diff(power_values).sum() == 0:
                            power_values = power_values[0]
                        else:
                            power_values = power_values.tolist()
                        dm[attribute] = power_values
                    else:
                        dm[attribute] = getattr(self, old_attribute)
                else:
                    if "cw" in old_attribute:
                        dm[attribute + "_start"] = None
                        dm[attribute + "_end"] = None
                    else:
                        dm[attribute] = None
        return pd.DataFrame.from_records([dm])

    def export_to_dataframe(self) -> pd.DataFrame:
        """returns a pd.DataFrame containing the appliance data

        Returns
        -------
        pd.DataFrame
            includes all the attributes and the user related information of an appliance.
        """
        return self.save()

    def __repr__(self):

        try:
            return self.save()[["user_name","num_users","name","number","power"]].to_string()
        except Exception:
            return ""


    def __eq__(self, other_appliance) -> bool:
        """checks the equality of two appliances

        Returns
        -------
        bool
            True if the two appliances:
                1. have the same attributes
                2. all their attributes have the same value
        """
        answer = np.array([])
        for attribute in APPLIANCE_ATTRIBUTES:
            if hasattr(self, attribute) and hasattr(other_appliance, attribute):
                np.append(
                    answer,
                    [getattr(self, attribute) == getattr(other_appliance, attribute)],
                )
            elif (
                hasattr(self, attribute) is False
                and hasattr(other_appliance, attribute) is False
            ):
                np.append(answer, True)
            else:
                if hasattr(self, attribute) is False:
                    print(f"{attribute} of appliance {self.name} is not assigned")
                else:
                    print(
                        f"{attribute} of appliance {other_appliance.name} is not assigned"
                    )
                np.append(answer, False)
        return answer.all()

    def windows(self, window_1:Iterable=None, window_2:Iterable=None,random_var_w:float=0 ,window_3:Iterable=None):
        """assings functioning windows to the appliance and adds the appliance to the user class

        Parameters
        ----------
        window_1 : Iterable, optional
            First functioning window, by default None

        window_2 : Iterable, optional
            Second functioning window, by default None

        window_3 : Iterable, optional
            Third functioning window, by default None

        random_var_w : Percentage, optional
            variability of the windows in percent, the same for all windows, by default 0

        Raises
        ------
        InvalidWindow

            * If number of specifies windows does not correspond to the given functioning windows.
            * If the sum of all windows time intervals for the appliance is smaller than the time the appliance is supposed to be on.

        Example
        --------
        If three time window is specified for the appliance as follow:

        #. from 00:00:00 to 00:20:00
        #. from 00:30:00 to 00:35:00
        #. from 00:40:00 to 00:55:00

        .. code-block:: python

            user.windows(
                window_1 = [0,20],
                window_2 = [30,35],
                window_3 = [40,55]
            )
        """

        if window_1 is None:
            warnings.warn(UserWarning("No windows is declared, default window of 24 hours is selected"))
            self.window_1 = np.array([0, 1440])
        else:
            self.window_1 = window_1

        if window_2 is None:
            if self.num_windows >= 2:
                raise InvalidWindow("Windows 2 is not provided although 2 windows were declared")
        else:
            self.window_2 = window_2

        if window_3 is None:
            if self.num_windows == 3:
                raise InvalidWindow("Windows 3 is not provided although 3 windows were declared")
        else:
            self.window_3 = window_3

        # check that the time allocated by the windows is larger or equal to the func_time of the appliance
        window_time = 0
        for i in range(1, self.num_windows + 1, 1):
            window_time = window_time + np.diff(getattr(self, f"window_{i}"))[0]
        if window_time < self.func_time:
            raise InvalidWindow(f"The sum of all windows time intervals for the appliance '{self.name}' of user '{self.user.user_name}' is smaller than the time the appliance is supposed to be on ({window_time} < {self.func_time}). Please check your input file for typos.")

        self.random_var_w = random_var_w
        self.daily_use = np.zeros(1440) #create an empty daily use profile
        self.daily_use[self.window_1[0]:(self.window_1[1])] = np.full(np.diff(self.window_1),0.001) #fills the daily use profile with infinitesimal values that are just used to identify the functioning windows
        self.daily_use[self.window_2[0]:(self.window_2[1])] = np.full(np.diff(self.window_2),0.001) #same as above for window2
        self.daily_use[self.window_3[0]:(self.window_3[1])] = np.full(np.diff(self.window_3),0.001) #same as above for window3

        self.random_var_1 = int(random_var_w*np.diff(self.window_1)) #calculate the random variability of window1, i.e. the maximum range of time they can be enlarged or shortened
        self.random_var_2 = int(random_var_w*np.diff(self.window_2)) #same as above
        self.random_var_3 = int(random_var_w*np.diff(self.window_3)) #same as above
        self.user.App_list.append(self) #automatically appends the appliance to the user's appliance list

        if self.fixed_cycle == 1:
            self.cw11 = self.window_1
            self.cw12 = self.window_2

    def assign_random_cycles(self):
        """
        Calculates randomised cycles taking the random variability in the duty cycle duration
        """
        if self.fixed_cycle >= 1:
            p_11 = random_variation(var=self.thermal_p_var, norm=self.p_11) #randomly variates the power of thermal apps, otherwise variability is 0
            p_12 = random_variation(var=self.thermal_p_var, norm=self.p_12) #randomly variates the power of thermal apps, otherwise variability is 0
            self.random_cycle1 = duty_cycle(var=self.r_c1, t1=self.t_11, p1=p_11, t2=self.t_12, p2=p_12) #randomise also the fixed cycle
            self.random_cycle2 = self.random_cycle1
            self.random_cycle3 = self.random_cycle1
            if self.fixed_cycle >= 2:
                p_21 = random_variation(var=self.thermal_p_var, norm=self.p_21) #randomly variates the power of thermal apps, otherwise variability is 0
                p_22 = random_variation(var=self.thermal_p_var, norm=self.p_22) #randomly variates the power of thermal apps, otherwise variability is 0
                self.random_cycle2 = duty_cycle(var=self.r_c2, t1=self.t_21, p1=p_21, t2=self.t_22, p2=p_22) #randomise also the fixed cycle

                if self.fixed_cycle >= 3:
                    p_31 = random_variation(var=self.thermal_p_var, norm=self.p_31) #randomly variates the power of thermal apps, otherwise variability is 0
                    p_32 = random_variation(var=self.thermal_p_var, norm=self.p_32) #randomly variates the power of thermal apps, otherwise variability is 0
                    self.random_cycle1 = random_choice(self.r_c1, t1=self.t_11, p1=p_11, t2=self.t_12, p2=p_12)

                    self.random_cycle2 = random_choice(self.r_c2, t1=self.t_21, p1=p_21, t2=self.t_22, p2=p_22)

                    self.random_cycle3 = random_choice(self.r_c3, t1=self.t_31, p1=p_31, t2=self.t_32, p2=p_32)

    def update_available_time_for_switch_on_events(self, indexes):
        """Remove the given time indexes from the ranges available to switch appliance on

        Parameters
        ----------
        indexes: list of int
            time indexes of the daily profile concerned by a new switch-on event

        Return
        ------
        nothing but can modify self.free_spots
        """
        # identify which of the unallocated time ranges contain the switch-on event
        spot_idx = None
        for i, fs in enumerate(self.free_spots):
            if indexes[0] >= fs.start and indexes[-1] <= fs.stop:
                spot_idx = i
                break
        if spot_idx is not None:
            spot_to_split = self.free_spots.pop(spot_idx)

            if indexes[0] == spot_to_split.start and indexes[-1] == spot_to_split.stop:
                pass  # nothing to do as the whole range should be removed, which is already the case from line above
            elif indexes[0] == spot_to_split.start:
                # reinsert a range going from end of indexes up to the end of picked range
                self.free_spots.insert(spot_idx, slice(indexes[-1] + 1, spot_to_split.stop, None))
            elif indexes[-1] == spot_to_split.stop:
                # reinsert a range going from beginning of picked range up to the beginning of indexes
                self.free_spots.insert(spot_idx, slice(spot_to_split.start, indexes[0], None))
            else:
                # split the range into 2 smaller ranges
                new_spot1 = slice(spot_to_split.start, indexes[0], None)
                new_spot2 = slice(indexes[-1] + 1, spot_to_split.stop, None)

                self.free_spots.insert(spot_idx, new_spot2)
                self.free_spots.insert(spot_idx, new_spot1)

    def update_daily_use(self, coincidence, power, indexes):
        """Update the daily use depending on existence of duty cycles of the Appliance instance

        This corresponds to step 2d. and 2e. of [1]

        [1] F. Lombardi, S. Balderrama, S. Quoilin, E. Colombo,
            Generating high-resolution multi-energy load profiles for remote areas with an open-source stochastic model,
            Energy, 2019, https://doi.org/10.1016/j.energy.2019.04.097.

        """

        if self.fixed_cycle > 0:  # evaluates if the app has some duty cycles to be considered
            evaluate = np.round(np.mean(indexes)) if indexes.size > 0 else 0
            # selects the proper duty cycle and puts the corresponding power values in the indexes range
            if evaluate in range(self.cw11[0], self.cw11[1]) or evaluate in range(self.cw12[0], self.cw12[1]):
                np.put(self.daily_use, indexes, (self.random_cycle1 * coincidence))
            elif evaluate in range(self.cw21[0], self.cw21[1]) or evaluate in range(self.cw22[0], self.cw22[1]):
                np.put(self.daily_use, indexes, (self.random_cycle2 * coincidence))
            else:
                np.put(self.daily_use, indexes, (self.random_cycle3 * coincidence))
        else:  # if no duty cycles are specified, a regular switch_on event is modelled
            # randomises also the App Power if thermal_p_var is on
            np.put(self.daily_use, indexes, (random_variation(var=self.thermal_p_var, norm=coincidence * power)))
        # updates the time ranges remaining for switch on events, excluding the current switch_on event
        self.update_available_time_for_switch_on_events(indexes)

    def calc_rand_window(self, window_idx=1, window_range_limits=[0, 1440]):
        _window = self.__getattribute__(f'window_{window_idx}')
        _random_var = self.__getattribute__(f'random_var_{window_idx}')
        rand_window = [random.randint(_window[0] - _random_var, _window[0] + _random_var),
                                random.randint(_window[1] - _random_var, _window[1] + _random_var)]
        if rand_window[0] < window_range_limits[0]:
            rand_window[0] = window_range_limits[0]
        if rand_window[1] > window_range_limits[1]:
            rand_window[1] = window_range_limits[1]

        return rand_window

    @property
    def maximum_profile(self):
        """Virtual maximum load profile of the appliance

        Returns
        --------
        np.array
            It assumes the appliance is always switched-on with maximum power and
            numerosity during all of its potential windows of use
        """
        return self.daily_use * np.mean(self.power) * self.number

    def specific_cycle(self, cycle_num, **kwargs):
        """assigining specific duty cycle for the appliace (maximum of three cycles can be assigned)

        Parameters
        ----------
        cycle_num : int
            represents the number of the specific cycle to be assigned. acceptable values are [1,2,3]

        **kwargs :
            additional features passed tp each specific cycle function. For example iff cycle_num = 1, **kwargs represents the arguments of function 'spefici_cycle_1' which are:
                * p_11
                * t_11
                * p_12
                * t_12
                * r_c1
                * cw11
                * cw12
        """
        if cycle_num == 1:
            self.specific_cycle_1(**kwargs)
        elif cycle_num == 2:
            self.specific_cycle_2(**kwargs)
        elif cycle_num == 3:
            self.specific_cycle_3(**kwargs)

    def specific_cycle_1(self, p_11 = 0, t_11 = 0, p_12 = 0, t_12 = 0, r_c1 = 0, cw11=None, cw12=None):
        """assigining the frist specific duty cycle for the appliace (maximum of three cycles can be assigned)

        Parameters
        ----------
        p_11 : float, optional
            Power rating for first part of first duty cycle. Only necessary if fixed_cycle is set to 1 or greater, by default 0

        t_11 : int[0,1440], optional
            Duration (minutes) of first part of first duty cycle. Only necessary if fixed_cycle is set to I or greater, by default 0

        p_12 : int, float, optional
            Power rating for second part of first duty cycle. Only necessary if fixed_cycle is set to 1 or greater, by default 0

        t_12 : int[0,1440], optional
            Duration (minutes) of second part of first duty cycle. Only necessary if fixed_cycle is set to I or greater, by default 0

        r_c1 : Percentage [0,1], optional
            randomization of the duty cycle parts duration. There will be a uniform random variation around t_i1 and t_i2. If this parameter is set to 0.1, then t_i1 and t_i2 will be randomly reassigned between 90% and 110% of their initial value; 0 means no randomisation, by default 0

        cw11 : Iterable, optional
            Window time range for the first part of first duty cycle number (not neccessarily linked to the overall time window), by default None

        cw12 : Iterable, optional
            Window time range for the first part of first duty cycle number (not neccessarily linked to the overall time window), by default None, by default None
        """
        self.p_11 = p_11
        self.t_11 = int(t_11)
        self.p_12 = p_12
        self.t_12 = int(t_12)
        self.r_c1 = r_c1
        if cw11 is not None:
            self.cw11 = cw11
        if cw12 is not None:
            self.cw12 = cw12
        # Below is not used
        self.fixed_cycle1 = np.concatenate(((np.ones(self.t_11)*p_11),(np.ones(self.t_12)*p_12))) #create numpy array representing the duty cycle

    def specific_cycle_2(self, p_21 = 0, t_21 = 0, p_22 = 0, t_22 = 0, r_c2 = 0, cw21=None, cw22=None):

        """assigining the frist specific duty cycle for the appliace (maximum of three cycles can be assigned)

        Parameters
        ----------
        p_21 : float, optional
            Power rating for first part of second duty cycle. Only necessary if fixed_cycle is set to 1 or greater, by default 0

        t_21 : int[0,1440], optional
            Duration (minutes) of first part of second duty cycle. Only necessary if fixed_cycle is set to I or greater, by default 0

        p_22 : int, float, optional
            Power rating for second part of second duty cycle. Only necessary if fixed_cycle is set to 1 or greater, by default 0

        t_22 : int[0,1440], optional
            Duration (minutes) of second part of second duty cycle. Only necessary if fixed_cycle is set to I or greater, by default 0

        r_c2 : Percentage [0,1], optional
            randomization of the duty cycle parts duration. There will be a uniform random variation around t_i1 and t_i2. If this parameter is set to 0.1, then t_i1 and t_i2 will be randomly reassigned between 90% and 110% of their initial value; 0 means no randomisation, by default 0

        cw21 : Iterable, optional
            Window time range for the first part of second duty cycle number (not neccessarily linked to the overall time window), by default None

        cw22 : Iterable, optional
            Window time range for the first part of second duty cycle number (not neccessarily linked to the overall time window), by default None, by default None
        """
        self.p_21 = p_21
        self.t_21 = int(t_21)
        self.p_22 = p_22
        self.t_22 = int(t_22)
        self.r_c2 = r_c2
        if cw21 is not None:
            self.cw21 = cw21
        if cw22 is not None:
            self.cw22 = cw22
        # Below is not used
        self.fixed_cycle2 = np.concatenate(((np.ones(self.t_21)*p_21),(np.ones(self.t_22)*p_22)))

    def specific_cycle_3(self, p_31 = 0, t_31 = 0, p_32 = 0, t_32 = 0, r_c3 = 0, cw31=None, cw32=None):
        """assigining the frist specific duty cycle for the appliace (maximum of three cycles can be assigned)

        Parameters
        ----------
        p_21 : float, optional
            Power rating for first part of third duty cycle. Only necessary if fixed_cycle is set to 1 or greater, by default 0

        t_21 : int[0,1440], optional
            Duration (minutes) of first part of third duty cycle. Only necessary if fixed_cycle is set to I or greater, by default 0

        p_22 : int, float, optional
            Power rating for second part of third duty cycle. Only necessary if fixed_cycle is set to 1 or greater, by default 0

        t_22 : int[0,1440], optional
            Duration (minutes) of second part of third duty cycle. Only necessary if fixed_cycle is set to I or greater, by default 0

        r_c2 : Percentage [0,1], optional
            randomization of the duty cycle parts duration. There will be a uniform random variation around t_i1 and t_i2. If this parameter is set to 0.1, then t_i1 and t_i2 will be randomly reassigned between 90% and 110% of their initial value; 0 means no randomisation, by default 0

        cw21 : Iterable, optional
            Window time range for the first part of third duty cycle number (not neccessarily linked to the overall time window), by default None

        cw22 : Iterable, optional
            Window time range for the first part of third duty cycle number (not neccessarily linked to the overall time window), by default None, by default None
        """
        self.p_31 = p_31
        self.t_31 = int(t_31)
        self.p_32 = p_32
        self.t_32 = int(t_32)
        self.r_c3 = r_c3
        if cw31 is not None:
            self.cw31 = cw31
        if cw32 is not None:
            self.cw32 = cw32
        # Below is not used
        self.fixed_cycle3 = np.concatenate(((np.ones(self.t_31)*p_31),(np.ones(self.t_32)*p_32)))

    #different time windows can be associated with different specific duty cycles
    def cycle_behaviour(self, cw11 = np.array([0,0]), cw12 = np.array([0,0]), cw21 = np.array([0,0]), cw22 = np.array([0,0]), cw31 = np.array([0,0]), cw32 = np.array([0,0])):
        """_summary_

        Parameters
        ----------
        cw11 : Iterable, optional
            Window time range for the first part of first duty cycle number, by default np.array([0,0])
        cw12 : Iterable, optional
            Window time range for the second part of first duty cycle number, by default np.array([0,0])
        cw21 : Iterable, optional
            Window time range for the first part of second duty cycle number, by default np.array([0,0])
        cw22 : Iterable, optional
            Window time range for the second part of second duty cycle number, by default np.array([0,0])
        cw31 : Iterable, optional
            Window time range for the first part of third duty cycle number, by default np.array([0,0])
        cw32 : Iterable, optional
            Window time range for the second part of third duty cycle number, by default np.array([0,0])
        """
        # only used around line 223
        self.cw11 = cw11 #first window associated with cycle1
        self.cw12 = cw12 #second window associated with cycle1
        self.cw21 = cw21 #same for cycle2
        self.cw22 = cw22
        self.cw31 = cw31 #same for cycle 3
        self.cw32 = cw32

    def rand_total_time_of_use(
        self,
        rand_window_1: Iterable[int],
        rand_window_2: Iterable[int],
        rand_window_3: Iterable[int],
        ) -> int:
        """Randomised total time of use of the Appliance instance
        """

        random_var_t = random_variation(var=self.time_fraction_random_variability)

        rand_time = round(
            random.uniform(self.func_time, int(self.func_time * random_var_t))
        )

        if rand_time < self.func_cycle:
            rand_time = self.func_cycle

        # total time available for appliance usage
        total_time = (
            (rand_window_1[1] - rand_window_1[0])
            + (rand_window_2[1] - rand_window_2[0])
            + (rand_window_3[1] - rand_window_3[0])
        )

        # check that the total randomised time of use does not exceed the total time available
        if rand_time > 0.99 * total_time:
            rand_time = int(0.99 * total_time)

        if rand_time < self.func_cycle:
            raise ValueError(f"The func_cycle you choose for appliance {self.name} might be too large to fit in the available time for appliance usage, please either reduce func_cycle or increase the windows of use of the appliance")
        return rand_time

    def rand_switch_on_window(self, rand_time:int):
        """Identifies a random switch on window within the available functioning windows

        This corresponds to step 2c. of:

            F. Lombardi, S. Balderrama, S. Quoilin, E. Colombo,
            Generating high-resolution multi-energy load profiles for remote areas with an open-source stochastic model,
            Energy, 2019, https://doi.org/10.1016/j.energy.2019.04.097.
        """

        indexes_choice = []
        for s in self.free_spots:
            if s.stop - s.start >= self.func_cycle:
                indexes_choice += [*range(s.start, s.stop - self.func_cycle + 1)] # this will be fast with cython
        n_choices = len(indexes_choice)
        if n_choices > 0:
            # Identifies a random switch on time within the available functioning windows
            # step 2c of [1]
            switch_on = indexes_choice[random.randint(0, n_choices-1)]
            spot_idx = None
            for i, fs in enumerate(self.free_spots):
                if fs.start <= switch_on <= fs.stop - self.func_cycle:
                    spot_idx = i
                    break

            largest_duration = min(rand_time, self.free_spots[spot_idx].stop - switch_on)

            if largest_duration > self.func_cycle:
                indexes = np.arange(switch_on, switch_on + (
                    int(random.uniform(self.func_cycle, largest_duration))))  # TODO randint
            elif largest_duration == self.func_cycle:
                indexes = np.arange(switch_on, switch_on + largest_duration)
            else:
                print("func time", self.func_cycle)
                print("max window", self.free_spots[spot_idx].stop)
                print("rand_time", rand_time)
                print("upper_limit", largest_duration)
                raise ValueError("There is something fishy with upper limit in switch on...")
        else:
            indexes = None
            # there are no available windows anymore

        return indexes

    def calc_coincident_switch_on(self, inside_peak_window:bool=True):
        """Computes how many of the 'n' Appliance instance are switched on simultaneously

        Implement eqs. 3 and 4 of [1]

        [1] F. Lombardi, S. Balderrama, S. Quoilin, E. Colombo,
            Generating high-resolution multi-energy load profiles for remote areas with an open-source stochastic model,
            Energy, 2019, https://doi.org/10.1016/j.energy.2019.04.097.
        """
        s_peak, mu_peak, op_factor = switch_on_parameters()

        # check if indexes are within peak window
        if inside_peak_window is True and self.fixed == 'no':
            # calculates coincident behaviour within the peak time range
            # eq. 4 of [1]
            coincidence = min(self.number, max(1, math.ceil(random.gauss(mu=(self.number * mu_peak + 0.5), sigma=(s_peak * self.number * mu_peak)))))
        # check if indexes are off-peak
        elif inside_peak_window is False and self.fixed == 'no':
            # calculates probability of coincident switch_ons off-peak
            # eq. 3 of [1]
            prob = random.uniform(0, (self.number - op_factor) / self.number)

            # randomly selects how many appliances are on at the same time
            array = np.arange(0, self.number) / self.number
            try:
                on_number = np.max(np.where(prob >= array)) + 1
            except ValueError:
                on_number = 1

            coincidence = on_number
        else:
            # All 'n' copies of an Appliance instance are switched on altogether
            coincidence = self.number
        return coincidence

    def generate_load_profile(self, prof_i, peak_time_range, day_type, power):
        """Generate load profile of the Appliance instance by updating its daily_use attribute

        Run steps 2a and 2b and repeat steps 2c – 2e of [1] until the sum of the durations of
        all the switch-on events equals the randomised total time of use of the Appliance

        [1] F. Lombardi, S. Balderrama, S. Quoilin, E. Colombo,
            Generating high-resolution multi-energy load profiles for remote areas with an open-source stochastic model,
            Energy, 2019, https://doi.org/10.1016/j.energy.2019.04.097.
        """
        # initialises variables for the cycle
        self.daily_use = np.zeros(1440)

        rand_daily_pref = 0 if self.user.user_preference == 0 else random.randint(1, self.user.user_preference)

        # skip this appliance in any of the following applies
        if (
                # evaluates if occasional use happens or not
                (random.uniform(0, 1) > self.occasional_use
                 # evaluates if daily preference coincides with the randomised daily preference number
                 or (self.pref_index != 0 and rand_daily_pref != self.pref_index)
                 # checks if the app is allowed in the given yearly behaviour pattern
                 or self.wd_we_type not in [day_type, 2])
        ):
            return

        # recalculate windows start and ending times randomly, based on the inputs
        rand_window_1 = self.calc_rand_window(window_idx=1)
        rand_window_2 = self.calc_rand_window(window_idx=2)
        rand_window_3 = self.calc_rand_window(window_idx=3)
        rand_windows = [rand_window_1, rand_window_2, rand_window_3]

        # random variability is applied to the total functioning time and to the duration
        # of the duty cycles provided they have been specified
        # step 2a of [1]
        rand_time = self.rand_total_time_of_use(rand_window_1, rand_window_2, rand_window_3)

        # redefines functioning windows based on the previous randomisation of the boundaries
        # step 2b of [1]
        if self.flat == 'yes':
            # for "flat" appliances the algorithm stops right after filling the newly
            # created windows without applying any further stochasticity
            total_power_value = self.power[prof_i] * self.number
            for rand_window in rand_windows:
                self.daily_use[rand_window[0]:rand_window[1]] = np.full(np.diff(rand_window),
                                                                       total_power_value)
            #single_load = single_load + self.daily_use
            return
        else:
            # "non-flat" appliances a mask is applied on the newly defined windows and
            # the algorithm goes further on
            for rand_window in rand_windows:
                self.daily_use[rand_window[0]:rand_window[1]] = np.full(np.diff(rand_window), 0.001)

        # calculates randomised cycles taking the random variability in the duty cycle duration
        self.assign_random_cycles()

        # steps 2c-2e repeated until the sum of the durations of all the switch-on events equals rand_time



        self.free_spots = [slice(rw[0], rw[1], None) for rw in rand_windows if rw[0] != rw[1]]

        tot_time = 0
        while tot_time <= rand_time:



            # one option could be to generate a lot of them at once
            indexes = self.rand_switch_on_window(
                rand_time=rand_time, #TODO maybe only consider rand_time-tot_time ...
            )
            if indexes is None:
                break # exit cycle and go to next Appliance as there are no available windows anymore


            # the count of total time is updated with the size of the indexes array
            tot_time = tot_time + indexes.size

            if tot_time > rand_time:
                # the total functioning time is reached, a correction is applied to avoid overflow of indexes
                indexes_adj = indexes[:-(tot_time - rand_time)]
                if len(indexes_adj) > 0:
                    inside_peak_window = within_peak_time_window(indexes_adj[0], indexes_adj[-1], peak_time_range[0], peak_time_range[-1])

                    # Computes how many of the 'n' of the Appliance instance are switched on simultaneously
                    coincidence = self.calc_coincident_switch_on(
                        inside_peak_window
                    )
                    # Update the daily use depending on existence of duty cycles of the Appliance instance
                    self.update_daily_use(
                        coincidence,
                        power=power,
                        indexes=indexes_adj
                    )
                break  # exit cycle and go to next Appliance

            else:
                inside_peak_window = within_peak_time_window(indexes[0], indexes[-1], peak_time_range[0], peak_time_range[-1])



                coincidence = self.calc_coincident_switch_on(
                    inside_peak_window
                )
                # Update the daily use depending on existence of duty cycles of the Appliance instance
                self.update_daily_use(
                    coincidence,
                    power=power,
                    indexes=indexes
                )






def switch_on_parameters():
    """
    Calibration parameters. These can be changed in case the user has some real data against which the model can be calibrated
    They regulate the probability of coincident switch-on within the peak window

    mu_peak corresponds to \mu_{%} in [1], p.8
    s_peak corresponds to \sigma_{%} in [1], p.8

    Notes
    -----
    [1] F. Lombardi, S. Balderrama, S. Quoilin, E. Colombo,
        Generating high-resolution multi-energy load profiles for remote areas with an open-source stochastic model,
        Energy, 2019, https://doi.org/10.1016/j.energy.2019.04.097.
    """

    mu_peak = 0.5  # median value of gaussian distribution [0,1] by which the number of coincident switch_ons is randomly selected
    s_peak = 0.5  # standard deviation (as percentage of the median value) of the gaussian distribution [0,1] above mentioned
    op_factor = 0.5  # off-peak coincidence calculation parameter

    return mu_peak, s_peak, op_factor


OLD_TO_NEW_MAPPING = {
    "name": "user_name",
    "n_users": "num_users",
    "us_pref": "user_preference",
    "P": "power",
    "P_11": "p_11",
    "P_12": "p_12",
    "P_21": "p_21",
    "P_22": "p_22",
    "P_31": "p_31",
    "P_32": "p_32",
    "pref_index": "pref_index",
    "P_series": "p_series",
    "Thermal_P_var": "thermal_p_var",
    "w": "num_windows",
    "cw11": "cw11",
    "cw12": "cw12",
    "cw21": "cw21",
    "cw22": "cw22",
    "cw31": "cw31",
    "cw32": "cw32",
    "fixed": "fixed",
    "flat": "flat",
    "c": "func_cycle",
    "t": "func_time",
    "n": "number",
    "occasional_use": "occasional_use",
    "r_c1": "r_c1",
    "r_c2": "r_c2",
    "r_c3": "r_c3",
    "r_t": "time_fraction_random_variability",
    "random_var_w": "random_var_w",
    "t_11": "t_11",
    "t_12": "t_12",
    "wd_we_type": "wd_we_type",
    "window_1": "window_1",
    "window_2": "window_2",
    "window_3": "window_3",
}

NEW_TO_OLD_MAPPING = {value: key for (key, value) in OLD_TO_NEW_MAPPING.items()}


APPLIANCE_ATTRIBUTES = (
            "name",
            "number",
            "power",
            "num_windows",
            "func_time",
            "time_fraction_random_variability",
            "func_cycle",
            "fixed",
            "fixed_cycle",
            "occasional_use",
            "flat",
            "thermal_p_var",
            "pref_index",
            "wd_we_type",
            "p_11",
            "t_11",
            "cw11",
            "p_12",
            "t_12",
            "cw12",
            "r_c1",
            "p_21",
            "t_21",
            "cw21",
            "p_22",
            "t_22",
            "cw22",
            "r_c2",
            "p_31",
            "t_31",
            "cw31",
            "p_32",
            "t_32",
            "cw32",
            "r_c3",
            "window_1",
            "window_2",
            "window_3",
            "random_var_w",
        )

APPLIANCE_ARGS = (
            "number",
            "power",
            #"p_series",
            "num_windows",
            "func_time",
            "time_fraction_random_variability",
            "func_cycle",
            "fixed",
            "fixed_cycle",
            "occasional_use",
            "flat",
            "thermal_p_var",
            "pref_index",
            "wd_we_type",
            "name",
        )

MAX_WINDOWS = 3
WINDOWS_PARAMETERS = ("window_1", "window_2", "window_3", "random_var_w")
DUTY_CYCLE_PARAMETERS = (
    ("p_11", "t_11", "cw11", "p_12", "t_12", "cw12", "r_c1"),
    ("p_21", "t_21", "cw21", "p_22", "t_22", "cw22", "r_c2"),
    ("p_31", "t_31", "cw31", "p_32", "t_32", "cw32", "r_c3"),
)



import os
import pandas as pd
import shutil

path = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__),
    )
)

available_data = {"shower":"shower_P.csv"}

def load_data(example:str) -> pd.DataFrame:

    if example not in available_data:
        raise ValueError(f"valid examples are {[*available_data]}")


    return pd.read_csv(
        f"{path}/{available_data[example]}"
    )


def download_example(destination:str):
    """ Copies the model files from the ramp package to a given path
    Parameters
    -----------
    destination : str
        The path to copy the model files.
    """
    files = ["input_file_1.py","input_file_2.py","input_file_3.py","shower_P.csv"]

    for file in files:
        shutil.copyfile(src=f"{path}/{file}", dst=f"{destination}/{file}")


#%% Import required modules

import sys, os, importlib

sys.path.append("../")
import argparse
from ramp.core.core import UseCase

parser = argparse.ArgumentParser(
    prog="python ramp_convert_old_input_files.py",
    description="Convert old python input files to xlsx ones",
)
parser.add_argument(
    "-i",
    dest="fname_path",
    nargs="+",
    type=str,
    help="path to the input file (including filename)",
)
parser.add_argument(
    "-o",
    dest="output_path",
    type=str,
    help="path where to save the converted filename",
)
parser.add_argument(
    "--suffix",
    dest="suffix",
    type=str,
    help="suffix appended to the converted filename",
    default=""
)


convert_names = """
# Code automatically added by ramp_convert_old_input_files.py
from ramp.core.core import Appliance
local_var_names = [(i, a) for i, a in locals().items() if isinstance(a, Appliance)]
for i, a in local_var_names:
    a.name = i
"""


def convert_old_user_input_file(
    fname_path, output_path=None, suffix="", keep_names=True
):
    """
    Imports an input file from a path and returns a processed User_list
    """

    line_to_change = -1

    # Check if the lines to save the names of the Appliance instances is already there
    # And check if the import of the User class is correct, otherwise adapt the file
    with open(fname_path, "r") as fp:
        lines = fp.readlines()
        if "# Code automatically added by ramp_convert_old_input_files.py\n" in lines:
            keep_names = False
        for i, l in enumerate(lines):
            if "from core import" in l:
                line_to_change = i
        # Change import statement by explicitly naming the full package path
        lines[line_to_change] = lines[line_to_change].replace("from core import", "from ramp.core.core import")

    # Modify import statement in file
    if line_to_change != -1:
        with open(fname_path, "w") as fp:
            fp.writelines(lines)

    # Add code to the input file to assign the variable name of the Appliance instances to their name attribute
    if keep_names is True:
        with open(fname_path, "a") as fp:
            fp.write(convert_names)

    fname = fname_path.replace(".py", "").replace(os.path.sep, ".")

    if output_path is None:
        output_path = os.path.dirname(fname_path)
    output_fname = fname_path.split(os.path.sep)[-1].replace(".py", suffix)
    output_fname = os.path.join(output_path, output_fname)

    if "ramp" not in fname:
        fname = f"ramp.{fname}"
    file_module = importlib.import_module(fname)

    user_list = file_module.User_list

    UseCase(users=user_list).save(output_fname)


if __name__ == "__main__":

    args = vars(parser.parse_args())
    fname = args["fname_path"]
    output_path = args.get("output_path")
    suffix = args.get("suffix")

    if fname is None:
        print("Please provide path to input file with option -i")
    else:
        if isinstance(fname, list):
            fnames = fname
            for fname in fnames:
                convert_old_user_input_file(
                    fname, output_path=output_path, suffix=suffix
                )
        else:
            convert_old_user_input_file(fname, output_path=output_path, suffix=suffix)



import json
import random
import time
import numpy as np
import pandas as pd
from openpyxl import load_workbook
from openpyxl.worksheet.cell_range import CellRange

POSSIBLE_FORMATS = """
    The possible formats of the power timeseries are :
        - a single value (int or float) if the power is constant throughout the load_profile (with random fluctuations if the variable 'thermal_p_var' is provided)
        - an array of value provided as text in a json array format, i.e. : [val1, val2, val3, ...]
        - a range of cells in another sheet of the input file (type '=' in the cell and then select the wished range of values to get the correct format automatically)

    ***Note***
    The last two formats will only accept one column (values only) or two columns (timestamps and values, respectively) and exactly 365 rows
"""


def read_input_file(filename):
    """Parse a RAMP .xlsx input file"""

    wb = load_workbook(filename=filename)
    sheet_names = wb.sheetnames
    name = sheet_names[0]
    headers = [c.value for c in wb[name][1]]
    df = pd.DataFrame(tuple(wb[name].values)[1:], columns=headers)
    df = df.fillna(value=np.nan)

    df["p_series"] = False
    for i, v in enumerate(df["power"].values):
        if isinstance(v, str):
            appliance_name = df["name"].iloc[i]
            user_name = df["user_name"].iloc[i]
            # the timeseries is provided as a range of values in the spreadsheet
            if "=" in v:
                ts_sheet_name, ts_range = v.replace("=", "").split("!")
                cr = CellRange(ts_range)
                if cr.size["columns"] == 1:
                    ts = json.dumps(
                        [
                            wb[ts_sheet_name].cell(row=c[0], column=c[1]).value
                            for c in cr.cells
                        ]
                    )
                elif cr.size["columns"] == 2:
                    ts = pd.DataFrame(
                        [
                            [
                                wb[ts_sheet_name].cell(row=c[0], column=c[1]).value
                                for c in col
                            ]
                            for col in cr.cols
                        ]
                    ).T
                    ts = ts.to_json(orient="values")
                else:
                    raise (
                        ValueError(
                            f"The provided range for the power timeseries of the appliance '{appliance_name}' of user '{user_name} spans more than two columns in '{filename}' (range {ts_range} of sheet '{ts_sheet_name}')\n{POSSIBLE_FORMATS}"
                        )
                    )
                if cr.size["rows"] != 365:
                    raise (
                        ValueError(
                            f"The provided range for the power timeseries of the appliance '{appliance_name}' of user '{user_name}' in '{filename}' does not contain 365 values as expected  (range {ts_range} of sheet '{ts_sheet_name}')\n{POSSIBLE_FORMATS}"
                        )
                    )
            # the timeseries is expected as an array in json format
            else:
                try:
                    ts = json.loads(v)
                    if len(ts) != 365:
                        raise (
                            ValueError(
                                f"The provided power timeseries of the appliance '{appliance_name}' of user '{user_name}' in '{filename}' does not contain 365 values as expected\n{POSSIBLE_FORMATS}"
                            )
                        )
                    ts = v
                except json.JSONDecodeError:
                    raise (
                        ValueError(
                            f"Could not parse the power timeseries provided for appliance '{appliance_name}' of user '{user_name}' in '{filename}'\n{POSSIBLE_FORMATS}"
                        )
                    )
            df.loc[i, "power"] = ts
            df.loc[i, "p_series"] = True
    return df


def random_variation(var, norm=1):
    """Pick a random variable within a uniform distribution of range [1-var, 1+var]

    Parameters
    ----------
    var: float
        sets the range of the uniform distribution around one
    norm: float
        multiplication factor of the random variable, default = 1

    Returns
    -------
    random number close to norm
    """
    return norm * random.uniform((1 - var), (1 + var))


def duty_cycle(var, t1, p1, t2, p2):
    """Assign a two period duty cycle

    concatenate an array where values equal p1 for a time (t1 +- random variation)
    followed by values equal to p2 for a time (t2 +- random variation)

    Parameters
    ----------
    var: float
        sets the range of the uniform distribution around t1 and t2
    t1: int
        time interval of the first part of the duty cycle in minutes
    p1: float
        power of the first part of the duty cycle in Watt
    t2: int
        time interval of the second part of the duty cycle in minutes
    p2: int
        power of the second part of the duty cycle in Watt

    Returns
    -------
    Power during each timestep of the duty cycle where p1 is repeated (t1 +- random variation) times and p2 is repeated (t2 +- random variation) times.
    The duty cycle is implicitly sampled every minutes (which is the unit for t1 and t2)
    """
    return np.concatenate(
        (
            np.ones(int(random_variation(var=-var, norm=t1))) * p1,
            np.ones(int(random_variation(var=-var, norm=t2))) * p2,
        )
    )


def random_choice(var, t1, p1, t2, p2):
    """Chooses one of two duty cycles randomly

    The choice is between a normal duty cycle and a reversed duty cycle (where t1 is swapped with t2 and p1 with p2)

    Parameters
    ----------
    var: float
        sets the range of the uniform distribution around t1 and t2
    t1: int
        time interval of the first part of the duty cycle in minutes
    p1: float
        power of the first part of the duty cycle in Watt
    t2: int
        time interval of the second part of the duty cycle in minutes
    p2: int
        power of the second part of the duty cycle in Watt

    Returns
    -------
    A duty cycle, see function duty_cycle
    """
    return random.choice(
        [
            duty_cycle(var, t1=t1, p1=p1, t2=t2, p2=p2),
            duty_cycle(var, t1=t2, p1=p2, t2=t1, p2=p1),
        ]
    )

def get_day_type(day):
    """Given a datetime object return 0 for weekdays or 1 for weekends"""
    if day.weekday() > 4:
        answer = 1
    else:
        answer = 0
    return answer


def yearly_pattern(year=None):
    """
    Definition of a yearly pattern of weekends and weekdays, in case some appliances have specific wd/we behaviour
    If no argument is provided, the pattern always starts a monday and lasts 365 days, otherwise the pattern matches
    the weekdays and weekends of the provided year
    """
    if year is None:
        year_behaviour = np.zeros(365)
        year_behaviour[5:365:7] = 1
        year_behaviour[6:365:7] = 1
        year_behaviour = year_behaviour.tolist()
    else:
        # a list with 0 for weekdays and 1 for weekends
        year_behaviour = pd.date_range(start=f"{year}-01-01", end=f"{year}-12-31", freq="D").map(get_day_type).to_list()
    return year_behaviour


def within_peak_time_window(win_start, win_stop, peak_win_start, peak_win_stop):
    """Given determines if a switch on window falls within the peak time window"""
    answer = True
    # start and stop of the given window are both below the lower limit of peak time window
    if win_start < peak_win_start and win_stop < peak_win_start:
        answer = False
    # start and stop of the given window are both above the upper limit of peak time window
    if win_start > peak_win_stop and win_stop > peak_win_stop:
        answer = False
    return answer


def calc_time_taken(func):
    """ Calculates the time elapsed during the execution of a function"""
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(func.__name__ + ' required ' + str((end-start)*1) + ' seconds for execution. ')
        return result
    return wrapper



# -*- coding: utf-8 -*-

#%% Definition of the inputs
'''
Input data definition
'''

from ramp.core.core import User
import pandas as pd
User_list = []


'''
This example input file represents a single household user whose only load
is the "shower". The example showcases how to model thermal loads by means of
the thermal_P_var attribute.
'''

#Create new user classes
HH = User("generic households",1)
User_list.append(HH)

HH_shower_P = pd.read_csv('example/shower_P.csv')

#High-Income
HH_shower = HH.Appliance(1,HH_shower_P,2,15,0.1,3, thermal_P_var = 0.2)
HH_shower.windows([390,540],[1080,1200],0.2)





# -*- coding: utf-8 -*-

#%% Import required libraries
import os.path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

BASE_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

#%% Post-processing
'''
Just some additional code lines to calculate useful indicators and generate plots
'''
def Profile_formatting(stoch_profiles):
    Profile_avg = np.zeros(1440)
    for pr in stoch_profiles:
        Profile_avg = Profile_avg + pr
    Profile_avg = Profile_avg/len(stoch_profiles)
    
    Profile_kW = []
    for kW in stoch_profiles:
        Profile_kW.append(kW/1000)
    
    Profile_series = np.array([])
    for iii in stoch_profiles:
        Profile_series = np.append(Profile_series,iii)
    
    return (Profile_avg, Profile_kW, Profile_series)

def Profile_cloud_plot(stoch_profiles,stoch_profiles_avg):
    #x = np.arange(0,1440,5)
    plt.figure(figsize=(10,5))
    for n in stoch_profiles:
        plt.plot(np.arange(1440),n,'#b0c4de')
        plt.xlabel('Time (hours)')
        plt.ylabel('Power (W)')
        plt.ylim(ymin=0)
        #plt.ylim(ymax=5000)
        plt.margins(x=0)
        plt.margins(y=0)
    plt.plot(np.arange(1440),stoch_profiles_avg,'#4169e1')
    plt.xticks([0,240,480,(60*12),(60*16),(60*20),(60*24)],[0,4,8,12,16,20,24])
    #plt.savefig('profiles.eps', format='eps', dpi=1000)
    plt.show()


def Profile_series_plot(stoch_profiles_series):
    #x = np.arange(0,1440,5)
    plt.figure(figsize=(10,5))
    plt.plot(np.arange(len(stoch_profiles_series)),stoch_profiles_series,'#4169e1')
    #plt.xlabel('Time (hours)')
    plt.ylabel('Power (W)')
    plt.ylim(ymin=0)
    #plt.ylim(ymax=5000)
    plt.margins(x=0)
    plt.margins(y=0)
    #plt.xticks([0,240,480,(60*12),(60*16),(60*20),(60*24)],[0,4,8,12,16,20,24])
    #plt.savefig('profiles.eps', format='eps', dpi=1000)
    plt.show()

#%% Export individual profiles
'''
for i in range (len(Profile)):
    np.save('p0%d.npy' % (i), Profile[i])
'''

# Export Profiles

def export_series(stoch_profiles_series, j=None, fname=None):
    series_frame = pd.DataFrame(stoch_profiles_series)
    if j is not None:
        series_frame.to_csv(os.path.join(BASE_PATH, 'results', 'output_file_%d.csv' % (j)))
    if fname is not None:
        series_frame.to_csv(os.path.join(BASE_PATH, 'results', f'output_file_{os.path.split(fname)[-1]}.csv'))



import argparse
import datetime
import os.path
import time

import pandas as pd
import numpy as np
from ramp.ramp_run import run_usecase

BASE_PATH = os.path.dirname(os.path.abspath(__file__))

parser = argparse.ArgumentParser(
    prog="python ramp_run.py", description="Execute RAMP code"
)
parser.add_argument(
    "-i",
    dest="fname_path",
    nargs="+",
    type=str,
    help="path to the (xlsx) input files (including filename). If not provided, then legacy .py input files will be fetched",
)
parser.add_argument(
    "-n",
    dest="num_profiles",
    nargs="+",
    type=int,
    help="number of profiles to be generated",
)

parser.add_argument(
    "-y",
    dest="years",
    nargs="+",
    type=int,
    help="Years for which one should generate demand profiles",
)

parser.add_argument(
    "--start-date",
    dest="date_start",
    type=datetime.date.fromisoformat,
    help="Date of start in YYYY-MM-DD format",
)

parser.add_argument(
    "--end-date",
    dest="date_end",
    type=datetime.date.fromisoformat,
    help="Date of end in YYYY-MM-DD format",
)

parser.add_argument(
    "--ext",
    dest="extension",
    type=str,
    help="Format of input files",
    default="xlsx"
)

parser.add_argument(
    "-p",
    dest="parallel",
    default=False,
    const=True,
    nargs="?",
    type=bool,
    help="Wether or not the simulation uses parallel processing",
)


def main():

    args = vars(parser.parse_args())
    fnames = args["fname_path"]
    num_profiles = args["num_profiles"]
    # Define which input files should be considered and run.
    date_start = args["date_start"]
    date_end = args["date_end"]
    ext = args["extension"]
    parallel_processing = args["parallel"]

    years = args["years"]

    if date_start is None:
        if date_end is not None:
            date_start = datetime.date(date_end.year, 1, 1)
    else:
        if date_end is None:
            date_end = datetime.date(date_start.year, 12, 31)

    month_files = False

    if years is not None:
        if date_start is not None or date_end is not None:
            raise ValueError("You cannot use the option -y in combinaison with --date-start and/or --date-end")
        else:
            date_start = datetime.date(years[0], 1, 1)
            date_end = datetime.date(years[-1], 12, 31)
        if len(years) == 1 and fnames is not None:
            # This special combination (option -y with only 1 year and option -i with a directory as parameter)
            # Triggers the special mode "one input file per month"
            if os.path.isdir(fnames[0]):
                dir_path = fnames[0]
                fnames = [os.path.join(dir_path, f) for f in os.listdir(fnames[0]) if f.endswith(ext)]
                fnames.sort(key=lambda f: int(''.join(filter(str.isdigit, f))))

                if len(fnames) == 12:
                    print("The following input files were found and will be used in this exact order for month inputs")
                    print("\n".join(fnames))
                    month_files = True
                    year = years[0]
                else:
                    raise ValueError(f"You want to simulate a whole year, yet the folder {dir_path} only contains {len(fnames)} out of the 12 monthes required")
            else:
                print("You selected a single year but the input path is not a folder.")


    if date_start is not None and date_end is not None:
        days = pd.date_range(start=date_start, end=date_end)
    else:
        days = None

    if fnames is None:
        print("Please provide path to input file with option -i, \n\nDefault to old version of RAMP input files\n")
        # Files are specified as numbers in a list (e.g. [1,2] will consider input_file_1.py and input_file_2.py)
        from ramp.ramp_run import input_files_to_run

        if num_profiles is not None:
            if len(num_profiles) == 1:
                num_profiles = num_profiles * len(input_files_to_run)
            else:
                if len(num_profiles) != len(input_files_to_run):
                    raise ValueError(
                        "The number of profiles parameters  should match the number of input files provided")
        else:
            num_profiles = [None] * len(input_files_to_run)

        for i, j in enumerate(input_files_to_run):
            run_usecase(j=j, num_profiles=num_profiles[i], parallel=parallel_processing)
    else:
        if num_profiles is not None:
            if len(num_profiles) == 1:
                num_profiles = num_profiles * len(fnames)
            else:
                if len(num_profiles) != len(fnames):
                    raise ValueError(
                        "The number of profiles parameters  should match the number of input files provided")
        else:
            num_profiles = [None] * len(fnames)
        if month_files is True:
            year_profile = []
            for i, fname in enumerate(fnames):
                month_start = datetime.date(year, i+1, 1)
                month_end = datetime.date(year, i+1, pd.Period(month_start, freq="D").days_in_month)
                days = pd.date_range(start=month_start, end=month_end, freq='D')
                monthly_profiles = run_usecase(fname=fname, num_profiles=num_profiles[i], days=days, plot=False, parallel=parallel_processing)
                year_profile.append(np.hstack(monthly_profiles))

            # Create a dataFrame to save the year profile with timestamps every minutes
            series_frame = pd.DataFrame(
                np.hstack(year_profile),
                index=pd.date_range(start=f"{year}-1-1", end=f"{year}-12-31 23:59", freq="T")
            )
            # Save to minute and hour resolution
            # TODO let the user choose where to save the files/file_name, make sure the user wants to overwrite the file
            # if it already exists
            series_frame.to_csv(os.path.join(BASE_PATH, 'yearly_profile_min_resolution.csv'))
            resampled = pd.DataFrame()

            resampled["mean"] = series_frame.resample("H").mean()
            resampled["max"] = series_frame.resample("H").max()
            resampled["min"] = series_frame.resample("H").min()
            #TODO add more columns with other resampled functions (do this in Jupyter)
            resampled.to_csv(os.path.join(BASE_PATH, 'yearly_profile_hourly_resolution.csv'))
        else:
            for i, fname in enumerate(fnames):
                run_usecase(fname=fname, num_profiles=num_profiles[i], days=days, parallel=parallel_processing)


if __name__ == "__main__":

    main()


# -*- coding: utf-8 -*-

#%% Definition of the inputs
'''
Input data definition
'''

from ramp.core.core import User

User_list = []


'''
This example input file represents a single household user whose only loads
are the "cooking" activities. The example showcases how to model electric cooking loads by means of
the Prefence Index and User Preference attributes.
'''

#Create new user classes
HH = User("generic household",1,3)
User_list.append(HH)

#Create new appliances

#Create Cooking appliances

HH_lunch1_soup = HH.Appliance(1,1800,2,70,0.15,60, thermal_P_var = 0.2, pref_index =1, fixed_cycle=1)
HH_lunch1_soup.windows([12*60,15*60],[0,0],0.15)
HH_lunch1_soup.specific_cycle_1(1800,10,750,60,0.15)
HH_lunch1_soup.cycle_behaviour([12*60,15*60],[0,0])

HH_lunch2_rice = HH.Appliance(1,1800,2,25,0.15,20, thermal_P_var = 0.2, pref_index = 2, fixed_cycle=1)
HH_lunch2_rice.windows([12*60,15*60],[0,0],0.15)
HH_lunch2_rice.specific_cycle_1(1800,10,750,15,0.15)
HH_lunch2_rice.cycle_behaviour([12*60,15*60],[0,0])

HH_lunch2_egg = HH.Appliance(1,1200,2,3,0.2,3, thermal_P_var = 0.2 , pref_index = 2)
HH_lunch2_egg.windows([12*60,15*60],[0,0],0.15)

HH_lunch2_platano = HH.Appliance(1,1800,2,10,0.15,5, thermal_P_var = 0.2, pref_index = 2, fixed_cycle=1)
HH_lunch2_platano.windows([12*60,15*60],[0,0],0.15)
HH_lunch2_platano.specific_cycle_1(1800,5,1200,5,0.15)
HH_lunch2_platano.cycle_behaviour([12*60,15*60],[0,0])

HH_lunch2_meat = HH.Appliance(1,1200,2,7,0.15,3, thermal_P_var = 0.2, pref_index = 2)
HH_lunch2_meat.windows([12*60,15*60],[0,0],0.15)

HH_lunch3_beansnrice = HH.Appliance(1,1800,2,45,0.2,30, thermal_P_var =0.2 , pref_index = 3, fixed_cycle=1)
HH_lunch3_beansnrice.windows([12*60,15*60],[0,0],0.15)
HH_lunch3_beansnrice.specific_cycle_1(1800,10,750,35,0.2)
HH_lunch3_beansnrice.cycle_behaviour([12*60,15*60],[0,0])

HH_lunch3_meat = HH.Appliance(1,1200,2,10,0.2,5, thermal_P_var = 0.2, pref_index = 3)
HH_lunch3_meat.windows([12*60,15*60],[0,0],0.15)

HH_lunch_yuca = HH.Appliance(1,1800,1,25,0.15,10, thermal_P_var = 0.2, pref_index =0, fixed_cycle=1)
HH_lunch_yuca.windows([13*60,14*60],[0,0],0.15)
HH_lunch_yuca.specific_cycle_1(1800,10,750,15,0.15)
HH_lunch_yuca.cycle_behaviour([12*60,15*60],[0,0])

HH_breakfast_huminta = HH.Appliance(1,1800,1,65,0.15,50, thermal_P_var = 0.2, pref_index =0, fixed_cycle=1)
HH_breakfast_huminta.windows([6*60,9*60],[0,0],0.15)
HH_breakfast_huminta.specific_cycle_1(1800,5,750,60,0.15)
HH_breakfast_huminta.cycle_behaviour([6*60,9*60],[0,0])

HH_breakfast_bread = HH.Appliance(1,1800,1,15,0.15,10, thermal_P_var = 0.2, pref_index =0, fixed_cycle=1)
HH_breakfast_bread.windows([6*60,9*60],[0,0],0.15)
HH_breakfast_bread.specific_cycle_1(1800,10,1200,5,0.15)
HH_breakfast_bread.cycle_behaviour([6*60,9*60],[0,0])

HH_breakfast_coffee = HH.Appliance(1,1800,1,5,0.15,2, thermal_P_var = 0.2, pref_index =0)
HH_breakfast_coffee.windows([6*60,9*60],[0,0],0.15)

HH_mate = HH.Appliance(1,1800,1,30,0.3,2, thermal_P_var = 0.2, pref_index =0)
HH_mate.windows([7*60,20*60],[0,0],0.15)





import os
import pytest
import numpy as np

from ramp.core.core import User, Appliance
from ramp.core.initialise import initialise_inputs
e
from ramp.ramp_convert_old_input_files import convert_old_user_input_file


def load_usecase(j=None, fname=None):
    peak_enlarge, user_list, num_profiles = initialise_inputs(
        j, fname, num_profiles=1
    )
    return user_list


class TestConversion:
    def setup_method(self):
        self.input_files_to_run = [1, 2, 3]
        self.file_suffix = "_test"
        os.chdir(
            "ramp"
        )  # for legacy code to work the loading of the input file has to happen from the ramp folder
        self.py_fnames = [
            os.path.join("input_files", f"input_file_{i}.py")
            for i in self.input_files_to_run
        ]
        self.xlsx_fnames = [
            os.path.join("test", f"input_file_{i}{self.file_suffix}.xlsx")
            for i in self.input_files_to_run
        ]
        for fname in self.xlsx_fnames:
            if os.path.exists(fname):
                os.remove(fname)

    def teardown_method(self):
        auto_code_proof = (
            "# Code automatically added by ramp_convert_old_input_files.py\n"
        )
        # remove created files
        for fname in self.xlsx_fnames:
            if os.path.exists(fname):
                os.remove(fname)
        # remove additional code in legacy input files to get the appliance name from python variable names
        for fname in self.py_fnames:
            with open(fname, "r") as fp:
                lines = fp.readlines()
                if auto_code_proof in lines:
                    idx = lines.index(auto_code_proof)
            with open(fname, "w") as fp:
                fp.writelines(lines[: idx - 1])

    def test_convert_py_to_xlsx(self):
        """Convert the 3 example .py input files to xlsx and compare each appliance of each user"""
        for i, j in enumerate(self.input_files_to_run):
            old_user_list = load_usecase(j=j)
            convert_old_user_input_file(
                self.py_fnames[i], output_path="test", suffix=self.file_suffix
            )
            new_user_list = load_usecase(fname=self.xlsx_fnames[i])
            for old_user, new_user in zip(old_user_list, new_user_list):
                if old_user != new_user:
                    pytest.fail()


def test_define_appliance_window_directly_equivalent_to_use_windows_method():
    user = User("test user", 1)

    params = dict(number=1, power=200, num_windows=1, func_time=0)
    win_start = 390
    win_stop = 540
    appliance1 = user.add_appliance(**params)
    appliance1.windows(window_1=[win_start, win_stop])

    params.update({"window_1": np.array([win_start, win_stop])})
    appliance2 = user.add_appliance(**params)

    assert appliance1 == appliance2


def test_define_appliance_duty_cycle_directly_equivalent_to_use_specific_cycle_method():
    user = User("test user", 1)

    params = dict(
        number=1,
        power=200,
        num_windows=1,
        func_time=0,
        window_1=[390, 540],
        fixed_cycle=1,
    )

    appliance1 = user.add_appliance(**params)
    cycle_params = {"p_11": 20, "t_11": 10, "cw11": np.array([400, 500])}
    appliance1.specific_cycle_1(**cycle_params)

    params.update(cycle_params)
    appliance2 = user.add_appliance(**params)

    assert appliance1 == appliance2


def test_provide_only_one_appliance_window_when_declaring_two():
    user = User("test user", 1)

    params = dict(number=1, power=200, num_windows=2, func_time=0)
    win_start = 390
    win_stop = 540
    with pytest.raises(ValueError):
        appliance1 = user.add_appliance(**params)
        appliance1.windows(window_1=[win_start, win_stop])
    with pytest.raises(ValueError):
        params.update({"window_1": np.array([win_start, win_stop])})
        user.add_appliance(**params)


def test_provide_no_appliance_window_when_declaring_one():
    user = User("test user", 1)

    params = dict(number=1, power=200, num_windows=1, func_time=0)
    with pytest.warns(UserWarning):
        appliance1 = user.add_appliance(**params)
        appliance1.windows()
    with pytest.warns(UserWarning):
        params.update({"window_1": None})
        user.add_appliance(**params)


def test_A():
    user = User("test user", 1)

    old_params = dict(n=1, P=200, w=1, t=0)
    win_start = 390
    win_stop = 540
    appliance1 = user.Appliance(user, **old_params)
    appliance1.windows(window_1=[win_start, win_stop])

    params = dict(number=1, power=200, num_windows=1, func_time=0, window_1=np.array([win_start, win_stop]))
    appliance2 = user.add_appliance(**params)

    assert appliance1 == appliance2

def test_B():
    user = User("test user", 1)

    params = dict(
        number=1,
        power=200,
        num_windows=1,
        func_time=0,
        window_1=[390, 540],
        fixed_cycle=1,
    )

    appliance1 = user.add_appliance(**params)
    cycle_params = {"p_11": 20, "t_11": 10, "cw11": np.array([400, 500])}
    appliance1.specific_cycle_1(**cycle_params)

    params.update(cycle_params)
    appliance2 = user.add_appliance(**params)

    assert appliance1 == appliance2


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import sphinx_pdj_theme

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath(".."))
sys.path.insert(0, os.path.abspath("../.."))
sys.path.insert(0, os.path.abspath("../../.."))

# -- Project information -----------------------------------------------------

project = 'RAMP'
copyright = '2022, Author List'
author = 'Author List'



# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "nbsphinx",
    "IPython.sphinxext.ipython_console_highlighting",
    "sphinx.ext.mathjax",
    "sphinx.ext.coverage",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx_copybutton',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
extensions.append("sphinxjp.themes.basicstrap")
html_theme = 'basicstrap'
#html_theme_path = [sphinx_pdj_theme.get_html_theme_path()]
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# copy btn settings
copybutton_prompt_text = "<AxesSubplot:>"




class InvalidType(TypeError):
    """Raises when an invalid data type is passed
    """
    pass

class InvalidWindow(ValueError):
    """Raises when an invalid functioning window is passed
    """
    pass




# -*- coding: utf-8 -*-

#%% Initialisation of a model instance

import numpy as np
import importlib
from ramp.core.core import UseCase



def user_defined_inputs(j=None, fname=None):
    """Imports an input file and returns a processed user_list

    Parameters
    ----------
    j: int
        the index of a .py input file that is in the format input_files/input_file_j.py
    fname: path
        path to a .xlsx input file
        if provided, overrides the loading of input_files/input_file_j.py

    Returns
    -------
    A list of User instances
    """

    # Back compatibility with old code
    if j is not None:
        file_module = importlib.import_module(f"ramp.example.input_file_{j}")
        user_list = file_module.User_list

    if fname is not None:
        usecase = UseCase()
        usecase.load(fname)
        user_list = usecase.users

    return user_list


def initialise_inputs(j=None, fname=None, num_profiles=None):
    """Loads the provided input file and prompt the user for number of profiles if not defined

    Parameters
    ----------
    j: int
        the index of a .py input file that is in the format input_files/input_file_j.py
    fname: path
        path to a .xlsx input file
        if provided, overrides the loading of input_files/input_file_j.py
    num_profiles: int
        the number of different usecase profiles which need to be generated

    Returns
    -------

    """

    user_list = user_defined_inputs(j, fname)

    peak_enlarge = 0.15  # percentage random enlargement or reduction of peak time range length, corresponds to \delta_{peak} in [1], p.7

    if num_profiles is None:
        # asks the user how many profiles (i.e. code runs) they want
        num_profiles = int(
            input("please indicate the number of profiles to be generated: ")
        )
        print("Please wait...")

    return (peak_enlarge, user_list, num_profiles)



# -*- coding: utf-8 -*-

#%% Import required libraries
import numpy as np
import random 
import math
from ramp.core.initialise import initialise_inputs
from ramp.core.core import UseCase

#%% Core model stochastic script


def calc_peak_time_range(user_list, peak_enlarge=0.15):
    """
    Calculate the peak time range, which is used to discriminate between off-peak and on-peak coincident switch-on probability
    Calculate first the overall Peak Window (taking into account all User classes).
    The peak time range corresponds to `peak time frame` variable in eq. (1) of [1]
    The peak window is just a time window in which coincident switch-on of multiple appliances assumes a higher probability than off-peak
    Within the peak window, a random peak time is calculated and then enlarged into a peak_time_range following again a random procedure

    Parameters
    ----------
    user_list: list
        list containing all the user types
    peak_enlarge: float
        percentage random enlargement or reduction of peak time range length
        corresponds to \delta_{peak} in [1], p.7

    Notes
    -----
    [1] F. Lombardi, S. Balderrama, S. Quoilin, E. Colombo,
        Generating high-resolution multi-energy load profiles for remote areas with an open-source stochastic model,
        Energy, 2019, https://doi.org/10.1016/j.energy.2019.04.097.

    Returns
    -------
    peak time range: numpy array
    """

    tot_max_profile = np.zeros(1440)  # creates an empty daily profile
    # Aggregate each User's theoretical max profile to the total theoretical max
    for user in user_list:
        tot_max_profile = tot_max_profile + user.maximum_profile
    # Find the peak window within the theoretical max profile
    peak_window = np.squeeze(np.argwhere(tot_max_profile == np.amax(tot_max_profile)))
    # Within the peak_window, randomly calculate the peak_time using a gaussian distribution
    peak_time = round(random.normalvariate(
        mu=round(np.average(peak_window)),
        sigma=1 / 3 * (peak_window[-1] - peak_window[0])
    ))
    rand_peak_enlarge = round(math.fabs(peak_time - random.gauss(mu=peak_time, sigma=peak_enlarge * peak_time)))
    # The peak_time is randomly enlarged based on the calibration parameter peak_enlarge
    return np.arange(peak_time - rand_peak_enlarge, peak_time + rand_peak_enlarge)


def stochastic_process(j=None, fname=None, num_profiles=None, day_type=None, parallel=False):
    """Generate num_profiles load profile for the usecase

        Covers steps 1. and 2. of the algorithm described in [1], p.6-7

    day_type: int
        0 for a week day or 1 for a weekend day

    Notes
    -----
    [1] F. Lombardi, S. Balderrama, S. Quoilin, E. Colombo,
        Generating high-resolution multi-energy load profiles for remote areas with an open-source stochastic model,
        Energy, 2019, https://doi.org/10.1016/j.energy.2019.04.097.
    """

    peak_enlarge, user_list, num_profiles = initialise_inputs(j, fname, num_profiles)

    # Calculation of the peak time range, which is used to discriminate between off-peak
    # and on-peak coincident switch-on probability, corresponds to step 1. of [1], p.6
    peak_time_range = calc_peak_time_range(user_list, peak_enlarge)

    uc = UseCase(users=user_list)

    if parallel is True:
        profiles = uc.generate_daily_load_profiles_parallel(num_profiles, peak_time_range, day_type)
    else:
        profiles = uc.generate_daily_load_profiles(num_profiles, peak_time_range, day_type)

    return profiles




from setuptools import find_packages,setup,Extension

exec(open("ramp/_version.py").read())


setup(
    name = "rampdemand",
    description= "An open-source python package for building bottom-up stochastic model for generating multi-energy load profiles",
    long_description = open("README.rst",encoding="utf8").read(),
    long_description_content_type = 'text/x-rst',
    author_email = "f.lombardi@tudelft.nl",
    url= "https://github.com/RAMP-project/RAMP",
    version = __version__,
    packages= find_packages(),
    license="European Union Public License 1.2",
    python_requires = ">=3.6.0",
    package_data={"": ["*.txt", "*.dat", "*.doc", "*.rst","*.xlsx"]},
    install_requires = [
        "pandas >= 1.3.3",
        "numpy >= 1.21.2",
        "xlsxwriter >= 1.3.7",
        "matplotlib >= 3.3.4",
        "openpyxl >= 3.0.6",
        "tqdm",
    ],
    # classifiers=[
    #     "Programming Language :: Python :: 3.7",
    #     "Programming Language :: Python :: 3.8",
    #     "Programming Language :: Python :: 3.9",
    #     "Intended Audience :: End Users/Desktop",
    #     "Intended Audience :: Developers",
    #     "Intended Audience :: Science/Research",
    #     "Operating System :: MacOS :: MacOS X",
    #     "Operating System :: Microsoft :: Windows",
    #     "Programming Language :: Python",
    #     "Topic :: Scientific/Engineering",
    # ],

    entry_points={
    "console_scripts": [
        "ramp=ramp.cli:main",
    ],
},

)



"""
RAMP: An open-source bottom-up stochastic model for generating multi-energy load profiles

RAMP is a bottom-up stochastic model for the generation of high-resolution multi-energy profiles,
conceived for application in contexts where only rough information about users' behaviour are obtainable.
Those may range from remote villages to whole countries.


Package dependencies:
    - numpy
    - matplotlib
    - math
    - random
"""


from ramp._version import __version__
from ramp.core.core import UseCase,User,Appliance
from ramp.core.stochastic_process import calc_peak_time_range
from ramp.core.utils import yearly_pattern
from ramp.example.examples import load_data,download_example

__authors__ = "Listed in AUTHORS"
__copyright__ = "Licensed under the European Union Public Licence (EUPL), Version 1.2-or-later"


